import asyncio
import heapq
import os
import pathlib
import urllib
import urllib.request

from enum import Enum, auto
from heapq import heappush, heappop
from typing import Union, Optional, Type, Dict, Tuple, List, Any
from urllib.parse import urlparse, ParseResult

from netaddr import IPAddress

from cyst.api.environment.clock import Clock
from cyst.api.environment.message import Resource as ResourceMsg, T, MessageType, Status, StatusValue, StatusOrigin
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.external import ExternalResources, Resource, ResourcePersistence, ResourceImpl
from cyst.api.host.service import Service, ActiveService
from cyst.api.logic.access import Authorization, AuthenticationToken, AuthenticationTarget
from cyst.api.logic.metadata import Metadata
from cyst.api.network.session import Session
from cyst.api.utils.counter import Counter


class ResourcesState(Enum):
    CREATED = auto()
    INIT = auto()
    OPENED = auto()
    CLOSED = auto()

class ResourceOp(Enum):
    SEND = auto()
    FETCH = auto()


# ----------------------------------------------------------------------------------------------------------------------
# Resource message is just a thin wrapper over resource contents to enable correct calling of process_message() of
# active services.
class ResourceMessage(ResourceMsg):
    def __init__(self, path: str, status: Status, data: str):
        self._path = path
        self._status = status
        self._data = data
        self._id = Counter().get("message")

    @property
    def type(self) -> MessageType:
        return MessageType.RESOURCE

    @property
    def id(self) -> int:
        return self._id

    @property
    def path(self) -> str:
        return self._path

    @property
    def status(self) -> Status:
        return self._status

    @property
    def data(self) -> Optional[str]:
        return self._data

    # ------------------------------------------------------------------------------------------------------------------
    # Unfortunately, these have to be implemented
    @property
    def src_ip(self) -> Optional[IPAddress]:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

    @property
    def dst_ip(self) -> Optional[IPAddress]:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

    @property
    def src_service(self) -> Optional[str]:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

    @property
    def dst_service(self) -> str:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

    @property
    def session(self) -> Optional[Session]:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

    @property
    def auth(self) -> Optional[Union[Authorization, AuthenticationToken, AuthenticationTarget]]:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

    @property
    def ttl(self) -> int:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

    @property
    def metadata(self) -> Metadata:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

    def set_metadata(self, metadata: Metadata) -> None:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

    @property
    def platform_specific(self) -> Dict[str, Any]:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

    def cast_to(self, type: Type[T]) -> T:
        raise NotImplementedError("Resources do not implement ordinary Message attributes.")

# ----------------------------------------------------------------------------------------------------------------------
# TODO: FileResource is internally synchronous. Which is not such a problem for intended use cases, but it should
#       probably be rewritten using aiofiles library.
class FileResource(ResourceImpl):

    def __init__(self):
        self._path: Optional[pathlib.Path] = None
        self._url = ""
        self._persistent = ResourcePersistence.TRANSIENT
        self._mode = "r"
        self._handle = None
        self._state = ResourcesState.CREATED

    @property
    def path(self) -> str:
        return self._url

    @property
    def persistence(self) -> ResourcePersistence:
        return self._persistent

    def init(self, path: ParseResult, params: Optional[dict[str, str]] = None, persistence: ResourcePersistence = ResourcePersistence.TRANSIENT) -> bool:
        # TODO: In the current incarnation, when filenames are supplied as a relative path then everything breaks down.
        self._url = str(path)
        self._path = urllib.request.url2pathname(path.path)
        self._persistent = persistence

        if params and "mode" in params:
            self._mode = params["mode"]

        self._state = ResourcesState.INIT

        return True

    def open(self) -> None:
        try:
            self._handle = open(self._path, self._mode)
        except Exception as e:
            raise RuntimeError(f"Failed to open a file with path '{self._path}' and mode '{self._mode}'. Reason: " + str(e))

        self._state = ResourcesState.OPENED

    def close(self) -> None:
        self._handle.close()
        self._state = ResourcesState.CLOSED

    async def send(self, data: str, params: Optional[dict[str, str]] = None) -> int:
        if ("w" in self._mode) or ("a" in self._mode):
            self._handle.write(data)
        else:
            raise RuntimeError("File not opened for writing")
        return len(data)

    async def receive(self, params: Optional[dict[str, str]] = None) -> str:
        if "r" in self._mode:
            data = self._handle.read()
            return data
        else:
            raise RuntimeError("File not opened for reading")


# ----------------------------------------------------------------------------------------------------------------------
class ExternalResourcesImpl(ExternalResources):
    def __init__(self, loop: asyncio.AbstractEventLoop, clock: Clock):
        self._loop = loop
        self._clock = clock

        self._tasks: Dict[str, Tuple[Union[asyncio.Future, ActiveService], asyncio.Task, Resource, float]] = {}
        self._finished: List[Tuple[float, int, Union[asyncio.Future, ActiveService], asyncio.Task, Resource]] = []
        self._pending: List[float] = []
        self._collecting = 0
        self._async_ops_running = 0

        self._resources: dict[str, Union[Type, ResourceImpl]] = {}
        self.register_resource("file", FileResource)

    @staticmethod
    def cast_from(o: ExternalResources) -> 'ExternalResourcesImpl':
        if isinstance(o, ExternalResourcesImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the ExternalResources interface")

    def register_resource(self, scheme: str, resource: Union[Type, ResourceImpl]) -> bool:
        if scheme in self._resources:
            raise RuntimeError(f"Attempting to register already registered resource for scheme {scheme}")
        else:
            if (isinstance(resource, type) and not issubclass(resource, ResourceImpl)) and not isinstance(resource, ResourceImpl):
                raise RuntimeError(f"Resource of wrong type (not implementing ResourceImpl) passed for registration: {resource}")
            self._resources[scheme] = resource
            return True

    def create_resource(self, path: str, params: Optional[dict[str, str]] = None, persistence: ResourcePersistence = ResourcePersistence.TRANSIENT) -> Resource:
        parsed_path = urlparse(path)

        if not parsed_path.scheme:
            raise RuntimeError(f"Resource path does not contain a scheme: {parsed_path}")

        if parsed_path.scheme not in self._resources:
            raise RuntimeError(f"Could not create a resource of unknown scheme {parsed_path.scheme}")

        resource_template = self._resources[parsed_path.scheme]
        if isinstance(resource_template, type):
            r: ResourceImpl = resource_template()
        else:
            r = resource_template
        r.init(parsed_path, params, persistence)  # Init should throw exception which gets caught by the Task

        if persistence == ResourcePersistence.PERSISTENT:
            r.open()

        return r

    def release_resource(self, resource: Resource) -> None:
        if resource.persistence == ResourcePersistence.PERSISTENT:
            if isinstance(resource, ResourceImpl):
                resource.close()

    async def send_async(self, resource: Union[str, Resource], data: str, params: Optional[dict[str, str]] = None, timeout: float = 0.0) -> int:
        r = resource
        if isinstance(resource, str):
            r = self.create_resource(resource, params)

        if not isinstance(r, ResourceImpl):
            raise RuntimeError("Malformed object passed with Resource interface.")

        if r.persistence == ResourcePersistence.TRANSIENT:
            r.open()

        self._async_ops_running += 1
        result = await self._schedule_task(r, ResourceOp.SEND, data, params, timeout)
        self._async_ops_running -= 1

        if r.persistence == ResourcePersistence.TRANSIENT:
            r.close()

        return result

    # Not really nice code duplication. But I still need to decide on whether jump into the rabbit hole and go full-on
    # on asyncio processing in the engine.
    def send(self, resource: Union[str, Resource], data: str, params: Optional[dict[str, str]] = None, timeout: float = 0.0, callback_service: Optional[ActiveService] = None) -> None:
        r = resource
        if isinstance(resource, str):
            r = self.create_resource(resource, params)

        if not isinstance(r, ResourceImpl):
            raise RuntimeError("Malformed object passed with Resource interface.")

        if r.persistence == ResourcePersistence.TRANSIENT:
            r.open()

        task = self._loop.create_task(r.send(data, params))
        self._tasks[task.get_name()] = (callback_service, task, r, self._clock.current_time() + timeout)
        task.add_done_callback(self._process_finished_tasks)
        heappush(self._pending, self._clock.current_time() + timeout)

    async def fetch_async(self, resource: Union[str, Resource], params: Optional[dict[str, str]] = None, timeout: float = 0.0) -> Optional[str]:
        r = resource
        if isinstance(resource, str):
            r = self.create_resource(resource, params)

        if not isinstance(r, ResourceImpl):
            raise RuntimeError("Malformed object passed with Resource interface.")

        if r.persistence == ResourcePersistence.TRANSIENT:
            r.open()

        self._async_ops_running += 1
        result = await self._schedule_task(r, ResourceOp.FETCH, "", params, timeout)
        self._async_ops_running -= 1

        if r.persistence == ResourcePersistence.TRANSIENT:
            r.close()

        return result

    def fetch(self, resource: Union[str, Resource], params: Optional[dict[str, str]] = None, timeout: float = 0.0, callback_service: Optional[ActiveService] = None) -> None:
        r = resource
        if isinstance(resource, str):
            r = self.create_resource(resource, params)

        if not isinstance(r, ResourceImpl):
            raise RuntimeError("Malformed object passed with Resource interface.")

        if r.persistence == ResourcePersistence.TRANSIENT:
            r.open()

        task = self._loop.create_task(r.receive(params))
        self._tasks[task.get_name()] = (callback_service, task, r, self._clock.current_time() + timeout)
        task.add_done_callback(self._process_finished_tasks)
        heappush(self._pending, self._clock.current_time() + timeout)

    async def _schedule_task(self, resource: ResourceImpl, resource_op: ResourceOp, data: str, params: Optional[dict[str, str]], timeout: float = 0.0) -> None | str | int:
        future = self._loop.create_future()

        if resource_op == ResourceOp.SEND:
            task = self._loop.create_task(resource.send(data, params))
        else:
            task = self._loop.create_task(resource.receive(params))

        real_timeout = self._clock.current_time() + timeout
        self._tasks[task.get_name()] = (future, task, resource, real_timeout)
        task.add_done_callback(self._process_finished_tasks)
        heappush(self._pending, real_timeout)

        # Tasks that fail timeout are cancelled and end here anyway
        await future
        # TODO: Exception not handled here!
        return future.result()

    def _process_finished_tasks(self, task: asyncio.Task):
        # Move task to finished
        future_or_service, _, resource, timeout = self._tasks[task.get_name()]
        heappush(self._finished, (timeout, Counter().get("resource"), future_or_service, task, resource))
        del self._tasks[task.get_name()]

        if resource.persistence == ResourcePersistence.TRANSIENT:
            if isinstance(resource, ResourceImpl):
                resource.close()

    def pending(self) -> Tuple[bool, float]:
        if not self._pending:
            return False, 0.0
        else:
            return True, self._pending[0] - self._clock.current_time()

    def collecting(self) -> bool:
        return self._collecting != 0

    def active(self) -> bool:
        return self.pending()[0] or self.collecting() or self._async_ops_running != 0

    def finish_collecting(self, task: asyncio.Task) -> None:
        self._collecting -= 1

    # Set futures on all finished tasks
    def collect_immediately(self) -> int:
        if not self._finished:
            return 0

        for timeout, _, future_or_service, task, resource in self._finished:
            self._pending.remove(timeout)
            if isinstance(future_or_service, asyncio.Future):
                if task.exception():
                    future_or_service.set_exception(task.exception())
                else:
                    future_or_service.set_result(task.result())
            else:
                # This is handling the case when a service callback is not specified, because we don't care (SEND case)
                if future_or_service:
                    if not task.exception():
                        status = Status(StatusOrigin.SYSTEM, StatusValue.SUCCESS)
                        t: asyncio.Task = self._loop.create_task(future_or_service.process_message(ResourceMessage(resource.path, status, task.result())))
                    else:
                        status = Status(StatusOrigin.SYSTEM, StatusValue.ERROR)
                        t: asyncio.Task = self._loop.create_task(future_or_service.process_message(ResourceMessage(resource.path, status, str(task.exception()))))
                    t.add_done_callback(self.finish_collecting)
                    self._collecting += 1

        # Reconstruct pending queue after reckless deletions
        heapq.heapify(self._pending)

        task_count = len(self._finished)
        self._finished.clear()
        return task_count

    # Set futures on tasks that past timeout
    def collect_at(self, current_time: float, ) -> int:
        if not self._finished:
            return 0

        task_count = 0

        timeout, _, _, _, _ = self._finished[0]
        while current_time <= timeout:
            self._pending.remove(timeout)
            _, _, future_or_service, task, resource = heappop(self._finished)
            task_count += 1
            if isinstance(future_or_service, asyncio.Future):
                future_or_service.set_result(task.result())
            else:
                if future_or_service:
                    if not task.exception():
                        status = Status(StatusOrigin.SYSTEM, StatusValue.SUCCESS)
                        t: asyncio.Task = self._loop.create_task(future_or_service.process_message(ResourceMessage(resource.path, status, task.result())))
                    else:
                        status = Status(StatusOrigin.SYSTEM, StatusValue.ERROR)
                        t: asyncio.Task = self._loop.create_task(future_or_service.process_message(ResourceMessage(resource.path, status, str(task.exception()))))
                    t.add_done_callback(self.finish_collecting)
                    self._collecting += 1

            if not self._finished:
                break

            timeout, _, _, _, _ = self._finished[0]

        # Reconstruct pending queue after reckless deletions
        heapq.heapify(self._pending)

        return task_count



import uuid

from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Optional, List, Tuple
from netaddr import IPAddress

from cyst.api.logic.data import Data


class AccessLevel(IntEnum):
    """ Access level represents in a simplified manner privileges of user, services, etc.

    Possible values:
        :NONE: No explicit access granted. This can either mean no access or guest-level access if available.
        :LIMITED: A user-level access.
        :ELEVATED: A root/superuser/administrator-level access
    """
    NONE = 0
    LIMITED = 1
    ELEVATED = 2


class AuthenticationTokenType(IntEnum):
    """ A type of authentication token.

    Possible values:
        :NONE: A sentinel value indicating misconfigured authentication token.
        :PASSWORD: A password-type auhentication token. Something that can be stored, shared, stolen, etc.
        :BIOMETRIC: A token that is bound to a user and cannot be appropriated unless the user is made to comply.
        :DEVICE: A physical token. Can be stolen by a human actor, but cannot be wholly manipulated by an agent.
    """
    NONE = 0,
    PASSWORD = 1,
    BIOMETRIC = 2,
    DEVICE = 3


class AuthenticationTokenSecurity(IntEnum):
    """ A security precautions on the authentication token.

    Possible values:
        :OPEN: There are no safeguards on the token. E.g, a plaintext password or any biometric.
        :SEALED: The token is hashed or encrypted for storage. E.g., /etc/shadow
        :HIDDEN: The token is never exposed. E.g., stored inside TPM.
    """
    OPEN = 0,
    SEALED = 1,
    HIDDEN = 2


class AuthenticationProviderType(IntEnum):
    """ A type of service providing an authentication.

    Possible values:
        :LOCAL: A service provides authentication for the node it resides on.
        :PROXY: A service provides authentication through another service residing on a remote node. E.g., openID.
        :REMOTE: A service provides authentication for another service on a remote node. E.g., sms or email codes.
    """
    LOCAL = 0,
    PROXY = 1,
    REMOTE = 2


class AuthenticationToken(ABC):
    """
    Authentication token represents a unit of information that can be used in one phase of authentication exchange
    to attempt to gain an authorization. It can be anything from PIN to biometrics.
    """

    @property
    @abstractmethod
    def type(self) -> AuthenticationTokenType:
        """
        Returns a type of the token

        :rtype: AuthenticationTokenType
        """

    @property
    @abstractmethod
    def security(self) -> AuthenticationTokenSecurity:
        """
        Returns a security measures of the token.

        :rtype: AuthenticationTokenSecurity
        """

    @property
    @abstractmethod
    def identity(self) -> str:
        """
        Returns or sets the identity of the token, i.e., who is associated with the token and which authorization would
        be gained if authenticated successfully.

        :getter: Returns the identity.
        :setter: Sets the identity. The token must enable setting the identity, otherwise False is returned.
        :type: str

        """

    @identity.setter
    def identity(self, value: str) -> bool:
        """
        Sets the identity of the token, i.e., who is associated with the token and which authorization would be
        gained if authenticated successfully.

        :param value: A new identity.
        :type value: str

        :return: Indication, whether the identity was correctly set.
        :rtype: bool
        """

    @abstractmethod
    def copy(self) -> Optional['AuthenticationToken']:
        """
        Creates a copy of the token.

        :rtype: AuthenticationToken
        """

    @property
    @abstractmethod
    def content(self) -> Optional[Data]:
        """
        Returns the data associated with the token. These data are mostly present with SEALED tokens, as they contained
        hashed/encrypted values that can be cracked before the token can be used.

        :rtype: Optional[Data]
        """


class AuthenticationTarget(ABC):
    """
    An authentication target represents on factor of an authentication scheme. It is a "pointer" to an existing
    service, with information about which tokens are accepted.
    """

    @property
    @abstractmethod
    def address(self) -> Optional[IPAddress]:
        """
        An IP address of the target. The address need not be present in case of local services.

        :rtype: Optional[IPAddress]
        """

    @property
    @abstractmethod
    def service(self) -> str:
        """
        A name of the service.

        :rtype: str
        """

    @property
    @abstractmethod
    def tokens(self) -> List[AuthenticationTokenType]:
        """
        A list of token types that the service is accepting.

        :rtype: List[AuthenticationTokenType]
        """


class Authorization(ABC):
    """
    Authorization represents a set of permissions that are in effect for the given combination of identity, services
    and nodes. Aside from identity, authorization does not enable inspecting the services and nodes it is working on.
    This information must be inferred from other channels.
    """

    @property
    @abstractmethod
    def identity(self) -> Optional[str]:
        """
        The identity of the owner of this authorization.
        TODO: That Optional[] is suspicious.

        :rtype: Optional[str]
        """

    @property
    @abstractmethod
    def access_level(self) -> AccessLevel:
        """
        The access level this authorization enables.

        :rtype: AccessLevel
        """

    @property
    @abstractmethod
    def expiration(self) -> int:
        """
        A number of simulated time units this authorization will be effective from the time it was created. If the
        value is -1, then the authorization never expires.

        :rtype:int
        """

    @property
    @abstractmethod
    def token(self) -> uuid.UUID:
        """
        A unique data token. This token can be used to compare and discern two or more authorizations.

        :rtype: uuid.UUID
        """


class AuthenticationProvider(ABC):
    """
    Authentication provider is a service that enables evaluation of authentication information. Depending on its type,
    it can authenticate only against itself (as a service), or can serve as an authenticator for other services.
    """

    @property
    @abstractmethod
    def type(self) -> AuthenticationProviderType:
        """
        The type of the provider.

        :rtype: AuthenticationProviderType
        """

    @property
    @abstractmethod
    def target(self) -> AuthenticationTarget:
        """
        An authentication target of this provider, i.e., the description of this provider in the form usable within the
        authentication framework.

        :rtype: AuthenticationTarget
        """

    @abstractmethod
    def token_is_registered(self, token: AuthenticationToken) -> bool:
        """
        Checks whether the supplied authentication token is registered in this provider.

        :param token: The token to check.
        :type token: AuthenticationToken

        :return: True if token is registered, false otherwise.
        """


class AccessScheme(ABC):
    """
    The access scheme represents the highest level of authentication/authorization framework. The scheme is a set of
    authentication factors (i.e., authentication providers) with associated authorizations. A scheme is linked to a
    service and describes the steps an outside actor must undertake to get access to its resources.
    """

    @property
    @abstractmethod
    def factors(self) -> List[Tuple[AuthenticationProvider, int]]:
        """
        Returns a list of authentication factors in form of authentication providers. For each factor it includes a
        time in simulated units that an actor has, before the authentication process is terminated.

        :rtype: List[Tuple[AuthenticationProvider, int]]:
        """

    @property
    @abstractmethod
    def identities(self) -> List[str]:
        """
        Returns a list of identities that are registered with the access scheme.

        :rtype: List[str]
        """

    @property
    @abstractmethod
    def authorizations(self) -> List[Authorization]:
        """
        Returns a list of authorization templates that are associated with the scheme. These are not actual
        authorizations. TODO: Are they? I am really not sure.

        :rtype: List[Authorization]
        """



from dataclasses import dataclass, field
from serde import serialize, coerce
from typing import Optional, List
from uuid import uuid4

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.configuration.infrastructure.log import LogConfig


@serialize(type_check=coerce)
@dataclass
class InfrastructureConfig(ConfigItem):
    """
    Infrastructure configuration serves for configuring the underlying machinery that powers CYST. As such, it is
    used for log setting, simulation modes, etc.

    :param log: Configuration of system logs.
    :type log: Optional[List[LogConfig]]
    """
    log: Optional[List[LogConfig]] = None
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__infrastructure"
    id: str = ""



from dataclasses import field
from datetime import datetime, time
from typing import List, Tuple, Dict, Any
from uuid import uuid4

from cyst.api.environment.configuration import PhysicalConfiguration
from cyst.api.environment.environment import Environment
from cyst.api.environment.physical import PhysicalAccess, PhysicalLocation, PhysicalConnection
from cyst.api.network.node import Node
from cyst.api.utils.duration import Duration
from cyst.core.environment.physical import PhysicalAccessImpl, PhysicalConnectionImpl, PhysicalLocationImpl
from cyst.platform.network.router import Router


class PhysicalConfigurationImpl(PhysicalConfiguration):
    def __init__(self, env: Environment):
        self.env = env
        self._locations: Dict[str, PhysicalLocation] = {}
        self._connections: Dict[Tuple[str, str], PhysicalConnection] = {}

    def create_physical_location(self, location_id: str | None) -> PhysicalLocation:
        location = PhysicalLocationImpl()
        if location_id is None:
            location_id = field(default_factory=lambda: str(uuid4()))
        self._locations[location_id] = location
        return location

    def get_physical_location(self, location_id: str) -> PhysicalLocation | None:
        return self._locations.get(location_id)

    def get_physical_locations(self) -> List[PhysicalLocation]:
        return list(self._locations.values())

    def remove_physical_location(self, location_id: str) -> None:
        if location_id not in self._locations:
            raise ValueError(f"Location '{location_id}' does not exist.")

        del self._locations[location_id]
        self._connections = {
            k: v for k, v in self._connections.items()
            if k[0] != location_id and k[1] != location_id
        }

    def create_physical_access(self, identity: str, time_from: datetime | None, time_to: datetime | None) -> PhysicalAccess:
        return PhysicalAccessImpl(identity=identity, time_from=time_from, time_to=time_to)

    def add_physical_access(self, location_id: str, access: PhysicalAccess) -> None:
        location = self.get_physical_location(location_id)
        if not location:
            raise ValueError(f"Location '{location_id}' does not exist.")
        location.access.append(access)

    def get_physical_accesses(self, location_id: str) -> List[PhysicalAccess]:
        location = self.get_physical_location(location_id)
        if not location:
            raise ValueError(f"Location '{location_id}' does not exist.")
        return location.access

    def remove_physical_access(self, location_id: str, access: PhysicalAccess) -> None:
        location = self.get_physical_location(location_id)
        if not location:
            raise ValueError(f"Location '{location_id}' does not exist.")
        location.access.remove(access)

    def add_physical_connection(self, origin: str, destination: str, travel_time: Duration) -> None:
        if origin not in self._locations:
            raise ValueError(f"Origin location '{origin}' does not exist.")
        if destination not in self._locations:
            raise ValueError(f"Destination location '{destination}' does not exist.")

        if (origin, destination) in self._connections or (destination, origin) in self._connections:
            return

        connection = PhysicalConnectionImpl(origin=origin, destination=destination, travel_time=travel_time)
        self._connections[(origin, destination)] = connection

    def remove_physical_connection(self, origin: str, destination: str) -> None:
        if origin not in self._locations:
            raise ValueError(f"Origin location '{origin}' does not exist.")
        if destination not in self._locations:
            raise ValueError(f"Destination location '{destination}' does not exist.")

        if (origin, destination) in self._connections:
            self._connections.pop((origin, destination), None)
        elif (destination, origin) in self._connections:
            self._connections.pop((destination, origin), None)


    def get_physical_connections(self, origin: str, destination: str | None) -> List[PhysicalConnection]:
        if origin not in self._locations:
            raise ValueError(f"Origin location '{origin}' does not exist.")
        if destination is not None and destination not in self._locations:
            raise ValueError(f"Destination location '{destination}' does not exist.")

        if destination:
            return [self._connections.get((origin, destination)) or self._connections.get((destination, origin))]
        return [
            conn for (orig, dest), conn in self._connections.items()
            if orig == origin or dest == origin
        ]

    def place_asset(self, location_id: str, asset: str) -> None:
        location = self.get_physical_location(location_id)
        if not location:
            raise ValueError(f"Location '{location_id}' does not exist.")

        current_location = self.get_location(asset)
        if current_location and current_location != location_id:
            self._locations[current_location].assets.remove(asset)
        elif current_location and current_location == location_id:
            return

        location.assets.append(asset)

    def remove_asset(self, location_id: str, asset: str) -> None:
        location = self.get_physical_location(location_id)
        if not location:
            raise ValueError(f"Location '{location_id}' does not exist.")

        if asset in location.assets:
            location.assets.remove(asset)

    def move_asset(self, origin: str, destination: str, asset: str) -> Tuple[bool, str, str]:
        if origin not in self._locations:
            raise ValueError(f"Origin location '{origin}' does not exist.")
        if destination not in self._locations:
            raise ValueError(f"Destination location '{destination}' does not exist.")

        if (origin, destination) not in self._connections and (destination, origin) not in self._connections:
            return False, origin, "No connection between locations"

        if asset not in self._locations[origin].assets:
            return False, "", "Asset not in origin location"

        try:
            self.env.configuration.general.get_object_by_id(asset, Node)
            asset_type = "Node"
        except KeyError or AttributeError:
            try:
                self.env.configuration.general.get_object_by_id(asset, Router)
                asset_type = "Router"
            except KeyError or AttributeError:
                asset_type = "User"

        # TODO determine if the asset is user by asking user manager, instead of this try-except block
        if asset_type == "User":
            access_rights = self.get_physical_accesses(destination)
            real_time = self.env.resources.clock.real_time().time()
            has_access = False

            for access in access_rights:
                identity_matches = access.identity in {asset, "*"}

                start_time = access.time_from.time() if access.time_from else time(0, 0, 0)
                end_time = access.time_to.time() if access.time_to else time(23, 59, 59)
                time_in_range = start_time <= real_time <= end_time

                if identity_matches and time_in_range:
                    has_access = True
                    break

            if not has_access:
                return False, origin, "User lacks access rights at the destination."

        self.remove_asset(origin, asset)
        self.place_asset(destination, asset)
        return True, destination, ""


    def get_assets(self, location_id: str) -> List[str]:
            location = self.get_physical_location(location_id)
            if not location:
                raise ValueError(f"Location '{location_id}' does not exist.")
            return location.assets

    def get_location(self, asset: str) -> str:
        for location_id, location in self._locations.items():
            if asset in location.assets:
                return location_id
        return ""

